/*
    Copyright 2004,2005,2006 Luigi Auriemma

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

    http://www.gnu.org/licenses/gpl.txt

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "md5.h"



#define TRUE    1
#define FALSE   0

#define VER     "0.1"
#define RETNULL { \
                *ret = 0; \
                return; \
                }
#define B64DECX bb = *bf; \
                for(; \
                bb && \
                ((bb == ' ') || \
                (bb == '\t') || \
                (bb == '\r') || \
                (bb == '\n')); \
                bf++);
#define BUFFSZ  1024
#ifndef u_char
    #define u_char  unsigned char
#endif



void EncodePassword_old(u_char *buf, u_char *ret, int retsize);
void EncodePassword(u_char *Key, u_char *Word, u_char *ret, int retsize, int decode);
u_char *Base64Decode(u_char *buf, u_char *ret);
void Base64Encode(u_char *buf, u_char *ret, int size);



int main(int argc, char *argv[]) {
    u_char  buff[BUFFSZ + 1];

    setbuf(stdout, NULL);

    fputs("\n"
        "nPOP saved password encoder/decoder "VER"\n"
        "by Luigi Auriemma\n"
        "e-mail: aluigi@autistici.org\n"
        "web:    aluigi.org\n"
        "\n", stdout);

    if(argc < 4) {
        printf("\nUsage: %s <dec/enc> <user> <password>\n"
            "\n"
            "Decoding/Encoding:\n"
            "0 = decode the stored password\n"
            "1 = encode a plain-text password\n"
            "\n", argv[0]);
        exit(1);
    }

    if(argv[1][0] & 1) {    // encoding
        EncodePassword(argv[2], argv[3], buff, BUFFSZ, TRUE);
        printf("\nEncoded password: %s\n\n", buff);
    } else {                // decoding
        EncodePassword(argv[2], argv[3], buff, BUFFSZ, FALSE);
        printf("\nDecoded password: %s\n\n", buff);
    }

    return(0);
}



void EncodePassword(u_char *Key, u_char *Word, u_char *ret, int retsize, int decode) {
    md5_context context;
    u_char      digest[16];
    u_char      *p,
                *r,
                *t;
    int         len,
                i;

    if(!Word || !*Word) RETNULL;

    if((decode == TRUE) && (*Word != '=')) {
        EncodePassword_old(Word, ret, retsize);
        return;
    }

    md5_starts(&context);
    md5_update(&context, Key, strlen(Key));
    md5_finish(&context, digest);

    p = strdup(Word);
    if(!p) RETNULL;
    len = strlen(p);

    if(decode == TRUE) {
        r = malloc(len + 1);
        if(!r) RETNULL;
        len = Base64Decode(p + 1, r) - r;
        free(p);
        p = r;
    }

    r = p;
    t = digest;
    for(i = 0; i < len; i++, r++, t++) {
        if(!(i & 0xf)) t = digest;
        *r ^= *t;
    }
    *r = 0;

    if(decode == TRUE) {
        strncpy(ret, p, retsize);
    } else {
        r = malloc((len << 1) + 5);
        if(!r) RETNULL;
        Base64Encode(p, r, len);
        free(p);
        p = r;

        *ret = '=';
        strncpy(ret + 1, p, retsize - 1);
    }
    free(p);
}



void EncodePassword_old(u_char *buf, u_char *ret, int retsize) {
    int     len;

    len = Base64Decode(buf, ret) - ret;
    while(len--) {
        *ret++ ^= 0xFF;
    }
    *ret = 0;
}



u_char *Base64Decode(u_char *buf, u_char *ret) {
    int     c,
            d,
            i,
            *Base;
    u_char  bb,
            *bf,
            *rf;

    if(!*buf) {
        *ret = 0;
        return(ret);
    }

    Base = (int *)malloc(sizeof(int) << 8);
    if(!Base) {
        strcpy(ret, buf);
        return(ret);
    }
    for(i = 0; i < 26; i++) {
        Base['A' + i] = i;
        Base['a' + i] = i + 26;
    }
    for(i = 0; i < 10; i++) {
        Base['0' + i] = i + 52;
    }
    Base['+'] = 62;
    Base['/'] = 63;

    bf = buf;
    rf = ret;
    B64DECX;
    while(1) {
        bb = *bf++;
        if((bb == '=') || !bb) break;
        c = Base[bb];

        B64DECX;
        bb = *bf++;
        if((bb == '=') || !bb) break;
        d = Base[bb];
        *rf++ = (c << 2) | (d >> 4);

        B64DECX;
        bb = *bf++;
        if((bb == '=') || !bb) break;
        c = Base[bb];
        *rf++ = ((d & 0xF) << 4) | (c >> 2);

        B64DECX;
        bb = *bf++;
        if((bb == '=') || !bb) break;
        *rf++ = ((c & 0x3) << 6) | Base[bb];
    }
    *rf = 0;
    free(Base);
    return(rf);
}



void Base64Encode(u_char *buf, u_char *ret, int size) {
    u_char      *r;
    int         i;
    const char  Base[] =
        "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        "abcdefghijklmnopqrstuvwxyz"
        "0123456789+/";

    i = 0;
    r = ret;
    while(1) {
        if((!size && !*buf) || ((size > 0) && (size <= i))) break;
        *r++ = Base[(*buf & 0xFC) >> 2];
        i++;
        buf++;

        if((!size && !*buf) || ((size > 0) && (size <= i))) {
            *buf = 0;
            *r++ = Base[((*(buf - 1) << 4) & 0x30) | ((*buf >> 4) & 0xF)];
            *r++ = '=';
            *r++ = '=';
            break;
        }
        *r++ = Base[((*(buf - 1) << 4) & 0x30) | ((*buf >> 4) & 0xF)];

        if((!size && !buf[i + 1]) || ((size > 0) && (size <= (i + 1)))) {
            *(buf + 1) = 0;
            *r++ = Base[((*buf << 2) & 0x3C) | ((*(buf + 1) >> 6) & 0x3)];
            *r++ = '=';
            break;
        }

        *r++ = Base[((*buf << 2) & 0x3C) | ((*(buf + 1) >> 6) & 0x3)];
        i++;
        buf++;

        *r++ = Base[*buf & 0x3F];
        i++;
        buf++;
    }
    *r = 0;
}


